from enum import EnumMeta

from graphene import Boolean, Enum, Float, ID, Int, List, String
from graphene.types.json import JSONString
from functools import singledispatch
from sqlalchemy import types
from sqlalchemy.dialects import postgresql

from .enums import enum_for_sa_enum
from .registry import get_global_registry

try:
    from sqlalchemy_utils import (
        ChoiceType,
        JSONType,
        ScalarListType,
        TSVectorType,
    )
except ImportError:
    ChoiceType = JSONType = ScalarListType = TSVectorType = object


@singledispatch
def convert_sqlalchemy_type(type_, column, registry=None):
    raise Exception(
        "Don't know how to convert the SQLAlchemy field %s (%s)"
        % (column, column.__class__)
    )


@convert_sqlalchemy_type.register(types.Time)
@convert_sqlalchemy_type.register(types.String)
@convert_sqlalchemy_type.register(types.Text)
@convert_sqlalchemy_type.register(types.Unicode)
@convert_sqlalchemy_type.register(types.UnicodeText)
@convert_sqlalchemy_type.register(postgresql.UUID)
@convert_sqlalchemy_type.register(postgresql.INET)
@convert_sqlalchemy_type.register(postgresql.CIDR)
@convert_sqlalchemy_type.register(postgresql.TSVECTOR)
@convert_sqlalchemy_type.register(TSVectorType)
def convert_column_to_string(type_, column, registry=None):
    return String


@convert_sqlalchemy_type.register(types.Date)
def convert_column_to_date(type_, column, registry=None):
    from graphene.types.datetime import Date

    return Date


@convert_sqlalchemy_type.register(types.DateTime)
def convert_column_to_datetime(type_, column, registry=None):
    from graphene.types.datetime import DateTime

    return DateTime


@convert_sqlalchemy_type.register(types.SmallInteger)
@convert_sqlalchemy_type.register(types.Integer)
def convert_column_to_int_or_id(type_, column, registry=None):
    return ID if column.primary_key else Int


@convert_sqlalchemy_type.register(types.Boolean)
def convert_column_to_boolean(type_, column, registry=None):
    return Boolean


@convert_sqlalchemy_type.register(types.Float)
@convert_sqlalchemy_type.register(types.Numeric)
@convert_sqlalchemy_type.register(types.BigInteger)
def convert_column_to_float(type_, column, registry=None):
    return Float


@convert_sqlalchemy_type.register(types.Enum)
def convert_enum_to_enum(type_, column, registry=None):
    return lambda: enum_for_sa_enum(type_, registry or get_global_registry())


# TODO Make ChoiceType conversion consistent with other enums
@convert_sqlalchemy_type.register(ChoiceType)
def convert_choice_to_enum(type_, column, registry=None):
    name = f"{column.table.name}_{column.name}".upper()
    if isinstance(type_.choices, EnumMeta):
        # type.choices may be Enum/IntEnum, in ChoiceType both presented as EnumMeta
        # do not use from_enum here because we can have more than one enum column in table
        return Enum(name, list((v.name, v.value) for v in type_.choices))
    else:
        return Enum(name, type_.choices)


@convert_sqlalchemy_type.register(ScalarListType)
def convert_scalar_list_to_list(type_, column, registry=None):
    return List(String)


@convert_sqlalchemy_type.register(types.ARRAY)
@convert_sqlalchemy_type.register(postgresql.ARRAY)
def convert_array_to_list(type_, column, registry=None):
    inner_type = convert_sqlalchemy_type(column.type.item_type, column)
    return List(inner_type)


@convert_sqlalchemy_type.register(postgresql.HSTORE)
@convert_sqlalchemy_type.register(postgresql.JSON)
@convert_sqlalchemy_type.register(postgresql.JSONB)
@convert_sqlalchemy_type.register(JSONType)
@convert_sqlalchemy_type.register(types.JSON)
def convert_json_type_to_string(type_, column, registry=None):
    return JSONString



import pkg_resources

from .fields import SQLAlchemyConnectionField
from .types import SQLAlchemyObjectType
from .utils import get_query
from .sql_mutation import (
    SQLAlchemyUpdateMutation,
    SQLAlchemyCreateMutation,
    SQLAlchemyDeleteMutation,
)

__version__ = "3.1.7"

__all__ = [
    "__version__",
    "SQLAlchemyObjectType",
    "SQLAlchemyConnectionField",
    "SQLAlchemyUpdateMutation",
    "SQLAlchemyCreateMutation",
    "SQLAlchemyDeleteMutation",
    "get_query",
]

if pkg_resources.get_distribution("SQLAlchemy").parsed_version.release < (1, 4):
    raise Exception("Use SQLAlchemy version > 1.4")



import graphene
import pytest
from graphene import Context

from alchql.fields import SQLAlchemyConnectionField
from alchql.middlewares import LoaderMiddleware
from alchql.node import AsyncNode
from alchql.sql_mutation import (
    SQLAlchemyCreateMutation,
    SQLAlchemyUpdateMutation,
)
from alchql.types import SQLAlchemyObjectType
from tests.models import Pet


@pytest.mark.asyncio
async def test_multiple_mutations(session):
    class PetType(SQLAlchemyObjectType):
        class Meta:
            model = Pet
            interfaces = (AsyncNode,)

    class MutationCreatePet(SQLAlchemyCreateMutation):
        class Meta:
            model = Pet
            output = PetType
            required_fields = ("name",)

    class MutationUpdatePet(SQLAlchemyUpdateMutation):
        class Meta:
            model = Pet
            output = PetType

    class MutationUpdatePetName(SQLAlchemyUpdateMutation):
        class Meta:
            model = Pet
            output = PetType
            only_fields = ["name"]

    class Query(graphene.ObjectType):
        node = AsyncNode.Field()
        all_pets = SQLAlchemyConnectionField(PetType.connection)

    class Mutation(graphene.ObjectType):
        create_pet = MutationCreatePet.Field()
        update_pet = MutationUpdatePet.Field()
        update_pet_name = MutationUpdatePetName.Field()

    schema = graphene.Schema(
        query=Query,
        mutation=Mutation,
    )

    query = """
        mutation CreatePet($value: MutationCreatePetInputType!) {
            createPet(value: $value) {
                id
                name
                petKind
            }
        }
    """

    result = await schema.execute_async(
        query,
        variables={
            "value": {
                "name": "asd",
                "petKind": "CAT",
                "hairKind": "SHORT",
            },
        },
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Pet]),
        ],
    )
    assert not result.errors
    pet_id = result.data["createPet"]["id"]

    query = """
        mutation UpdatePet($value: MutationUpdatePetInputType!, $updatePetId: ID!) {
            updatePet(value: $value, id: $updatePetId) {
                id
                name
            }
        }
    """

    result = await schema.execute_async(
        query,
        variables={
            "updatePetId": pet_id,
            "value": {
                "name": "dsa",
                "hairKind": "LONG",
            },
        },
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Pet]),
        ],
    )
    assert not result.errors
    assert result.data["updatePet"]["name"] == "dsa"

    query = """
        mutation UpdatePet($value: MutationUpdatePetNameInputType!, $updatePetId: ID!) {
            updatePetName(value: $value, id: $updatePetId) {
                id
                name
            }
        }
    """

    result = await schema.execute_async(
        query,
        variables={
            "updatePetId": pet_id,
            "value": {
                "name": "eeee",
            },
        },
        context_value=Context(session=session),
        middleware=[
            LoaderMiddleware([Pet]),
        ],
    )
    assert not result.errors
    assert result.data["updatePetName"]["name"] == "eeee"



import re

from graphene import Dynamic, Field, ResolveInfo
from graphene.types.objecttype import ObjectTypeMeta
from sqlalchemy import ForeignKey
from sqlalchemy.orm import RelationshipProperty

from .gql_fields import camel_to_snake


def set_object_type(root, info: ResolveInfo):
    field_name = info.field_name
    root_type = type(root)
    if not hasattr(root_type, "_meta"):
        return
    types = getattr(info.context, "object_types", {})
    if types.get(info.field_name) is not None:
        return

    connection_field_type = root_type._meta.fields[camel_to_snake(field_name)]
    type_ = None
    if isinstance(connection_field_type, Field):
        type_ = connection_field_type.type
    elif isinstance(connection_field_type, Dynamic):
        if isinstance(connection_field_type.type(), ObjectTypeMeta):
            type_ = connection_field_type.type().type
        elif type(connection_field_type.type()) == Field:
            type_ = connection_field_type.type().type

    if not type_:
        return

    types[info.field_name] = type_
    setattr(info.context, "object_types", types)


def get_batch_resolver(relationship_prop: RelationshipProperty, single=False):
    async def resolve(root, info: ResolveInfo, **args):
        key = (
            relationship_prop.parent.entity,
            relationship_prop.mapper.entity,
            relationship_prop.key,
        )
        _loader = info.context.loaders[key]

        _loader.info = info
        set_object_type(root, info)

        key = getattr(root, next(iter(relationship_prop.local_columns)).key)
        if not key:
            p = None
        else:
            p = await _loader.load(key)

        if single:
            return p[0] if p else None

        return p

    return resolve


def get_fk_resolver(fk: ForeignKey, single=False):
    async def resolve(root, info: ResolveInfo, **args):
        key = (
            fk.constraint.table,
            fk.constraint.referred_table,
            re.sub(r"_(?:id|pk)$", "", fk.parent.key),
        )
        _loader = info.context.loaders[key]

        _loader.info = info
        set_object_type(root, info)

        key = getattr(root, fk.parent.key)
        if not key:
            p = None
        else:
            p = await _loader.load(key)

        if single:
            return p[0] if p else None

        return p

    return resolve


def get_fk_resolver_reverse(fk: ForeignKey, single=False):
    async def resolve(root, info: ResolveInfo, **args):
        key = (
            fk.constraint.referred_table,
            fk.constraint.table,
            str(fk.constraint.table.fullname),
        )
        _loader = info.context.loaders[key]

        _loader.info = info
        set_object_type(root, info)

        key = getattr(root, fk.column.key)
        if not key:
            p = None
        else:
            p = await _loader.load(key)

        if single:
            return p[0] if p else None

        return p

    return resolve



from typing import Optional, Type

from graphene import Connection, PageInfo

from .utils import (
    get_offset_with_default,
    offset_to_cursor,
)


def connection_from_array_slice(
    array_slice,
    args: Optional[dict] = None,
    slice_start: int = 0,
    array_length: Optional[int] = None,
    array_slice_length: Optional[int] = None,
    connection_type: Type[Connection] = Connection,
) -> Connection:
    """Create a connection object from a slice of the result set.

    Note that different from its JavaScript counterpart which expects an array,
    this function accepts any kind of sliceable object. This object represents
    a slice of the full result set. You need to pass the start position of the
    slice as `slice start` and the length of the full result set as `array_length`.
    If the `array_slice` does not have a length, you need to provide it separately
    in `array_slice_length` as well.

    This function is similar to `connection_from_array`, but is intended for use
    cases where you know the cardinality of the connection, consider it too large
    to materialize the entire result set, and instead wish to pass in only a slice
    of the total result large enough to cover the range specified in `args`.

    If you do not provide a `slice_start`, we assume that the slice starts at
    the beginning of the result set, and if you do not provide an `array_length`,
    we assume that the slice ends at the end of the result set.
    """
    args = args or {}
    edge_type = connection_type.Edge
    page_info_type = getattr(connection_type, "PageInfo", PageInfo)

    before = args.get("before")
    after = args.get("after")
    first = args.get("first")
    last = args.get("last")

    if array_slice_length is None:
        array_slice_length = len(array_slice)
    slice_end = slice_start + array_slice_length
    if array_length is None:
        array_length = slice_end

    start_offset = max(slice_start, 0)
    end_offset = min(slice_end, array_length)

    after_offset = get_offset_with_default(after, -1)
    if 0 <= after_offset < array_length:
        start_offset = max(start_offset, after_offset + 1)

    before_offset = get_offset_with_default(before, end_offset)
    if 0 <= before_offset < array_length:
        end_offset = min(end_offset, before_offset)

    if isinstance(first, int):
        if first < 0:
            raise ValueError("Argument 'first' must be a non-negative integer.")

        end_offset = min(end_offset, start_offset + first)
    if isinstance(last, int):
        if last < 0:
            raise ValueError("Argument 'last' must be a non-negative integer.")

        start_offset = max(start_offset, end_offset - last)

    # If supplied slice is too large, trim it down before mapping over it.
    trimmed_slice = array_slice[start_offset - slice_start : end_offset - slice_start]

    edges = [
        edge_type(node=value, cursor=offset_to_cursor(start_offset + index))
        for index, value in enumerate(trimmed_slice)
    ]

    first_edge_cursor = edges[0].cursor if edges else None
    last_edge_cursor = edges[-1].cursor if edges else None
    lower_bound = after_offset + 1 if after else 0
    upper_bound = before_offset if before else array_length

    return connection_type(
        edges=edges,
        page_info=page_info_type(
            start_cursor=first_edge_cursor,
            end_cursor=last_edge_cursor,
            has_previous_page=isinstance(last, int) and start_offset > lower_bound,
            has_next_page=isinstance(first, int) and end_offset < upper_bound,
        ),
    )
